// point-location-query
// Read a planar subdivision from a file, answer point location queries
// MCS 481 project 3 description version 1.0
#include <CGAL/basic.h>
#include <CGAL/version.h>
#include <CGAL/Cartesian.h>
#include <CGAL/MP_Float.h>
#include <CGAL/Arr_segment_traits_2.h>
#include <CGAL/Arrangement_2.h>
#include <CGAL/Arr_trapezoid_ric_point_location.h>

#if CGAL_VERSION_NR >= CGAL_VERSION_NUMBER(3,4,0)
  #define INSERT CGAL::insert
#else
  #define INSERT CGAL::insert_x_monotone_curves
#endif

#include <iostream>
#include <sstream>
#include <list>
#include <vector>
#include <string>
#include <ctime>
using namespace std;

typedef CGAL::Quotient<CGAL::MP_Float>    Number_type;
typedef CGAL::Cartesian<Number_type>      K;
typedef CGAL::Arr_segment_traits_2<K>     Traits;
typedef Traits::Point_2                   Point;
typedef Traits::X_monotone_curve_2        Segment;
typedef CGAL::Arrangement_2<Traits>       Arrangement;

typedef CGAL::Arr_trapezoid_ric_point_location<Arrangement> PointLocator;

// CGAL SUPPORTS OTHER POINT LOCATION METHODS

// Naive - defined in CGAL/Arr_naive_point_location.h
// typedef CGAL::Arr_naive_point_location<Arrangement>         Naive_pl;

// Landmarks - defined in CGAL/Arr_landmarks_point_location.h
// typedef CGAL::Arr_landmarks_point_location<Arrangement>     Landmarks_pl;

// global time variable
clock_t LAST;

void tic()
{
  LAST = clock();
}

void toc(string s)
{
  double t = (double)(clock() - LAST) / (double)CLOCKS_PER_SEC;
  cerr << "TIMING: " << t;
  if (s.size() > 0) {
    cerr << " " << s << endl;
  } else {
    cerr << endl;
  }
}

int main(int argc, char *argv[])
{
  vector<Point> verts;
  list<Segment> segs;

  if (argc < 2) {
    cout << "Usage: " << argv[0] << " SUBDIV_FILE [QUERY_FILE]" << endl;
    cout << "If no QUERY_FILE specified, queries are read from STDIN." << endl;
    exit(-1);
  }

  //------------------------------------------------------------
  // PARSE INPUT
  //------------------------------------------------------------

  std::ifstream subfile;
  subfile.open(argv[1]);
  if (subfile.fail()) {
    cerr << "Unable to open \"" << argv[1] << "\"." << endl;
    exit(-1);
  }

  // READ VERTICES, stop at blank line
  int n = 0;
  Point p;
  string line;
  while (getline(subfile,line)) {
    n++;
    if (line.size() == 0) // blank line
      break;
    if (line[0] == '#')  // comment
      continue;
    istringstream linestream(line);
    if (!(linestream >> p)) {
      // EXIT due to bad input
      cerr << "Bad input (vertex) at line " << n
           << ": \"" << line << "\"" << endl;
      exit(-1);
    }
    verts.push_back(p);
  }
  cerr << "Read " << verts.size() << " vertices." << endl;

  // READ EDGES, stop at eof
  while (getline(subfile,line)) {
    n++;
    if (line.size() == 0) // blank line
      continue;
    if (line[0] == '#')  // comment
      continue;
    istringstream linestream(line);
    int istart,iend;
    if (!(linestream >> istart >> iend)) {
      // EXIT due to bad input
      cerr << "Bad input (edge) at line " << n << ": \"" << line << "\"" << endl;
      exit(-1);
    }
    if ((istart < 0) || (istart >= (int)verts.size()) ||
        (iend < 0) || (iend >= (int)verts.size())) {
      // EXIT due to invalid index
      cerr << "Index out of bounds (nvert=" << verts.size()
                << ") on line " << n
                << ": \"" << line << "\"" << endl;
      exit(-1);
    }
    Segment s(verts[istart],verts[iend]);
    segs.push_back(s);
  }
  cerr << "Read " << segs.size() << " edges." << endl;
  subfile.close();

  //------------------------------------------------------------
  // CREATE ARRANGMENT & SEARCH STRUCTURE
  //------------------------------------------------------------

  Arrangement arr;

  tic();
  INSERT(arr,segs.begin(),segs.end());
  toc("seconds - build DCEL");

  PointLocator D;

  tic();
  D.attach(arr);
  toc("seconds - build search structure");

  //------------------------------------------------------------
  // PERFORM QUERIES
  //------------------------------------------------------------
  std::istream *qin;
  std::ifstream qfile;

  if (argc < 3) {
    qin = &cin;
  } else {
    qfile.open(argv[2]);
    if (qfile.fail()) {
      cerr << "Unable to open \"" << argv[2] << "\"." << endl;
      exit(-1);
    }
    qin = &qfile;
  }

  tic();
  Point q;
  n = 0;
  while (*qin >> q) {
    // This is the actual query
    CGAL::Object obj = D.locate(q);

    // Return type is an object from the DCEL that contains q
    // Now we figure out what kind of object (face or not) and report it
    cout << "Query " << n++ << ": ("
         << CGAL::to_double(q.x()) << "," << CGAL::to_double(q.y())
         << ") ";
    Arrangement::Face_const_handle f;
    if (CGAL::assign(f,obj)) {
      // Now f is the handle of a face containing q
      cout << "is located in ";
      if (f->is_unbounded())
        cout << "the unbounded face ";
      else
        cout << "a bounded face ";
      cout << "(which has " << f->number_of_holes() << " holes)" << endl;
    } else {
      // Object containing q is not a face
      cout << "lies on an edge or vertex of the subdivision." << endl;
    }
  }
  // Timing less reliable here, since it also counts IO overhead
  toc("seconds - queries");

  if (qfile.is_open())
    qfile.close();

  return 0;
}
